/*
**  Copyright 2001-2003  University of Illinois Board of Trustees
**  Copyright 2001-2005  Mark D. Roth
**  All rights reserved. 
**
**  pam_per_user.c - PAM module for calling per-user authentication mechanisms
**
**  Mark D. Roth <roth@feep.net>
*/

#include <config.h>

#include <stdio.h>
#include <syslog.h>
#include <errno.h>
#include <grp.h>
#include <pwd.h>

#ifdef STDC_HEADERS
# include <string.h>
# include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#include <security/pam_appl.h>
#include <security/pam_modules.h>

#include <compat.h>


#define PAM_PER_USER_DEFAULT_MAPFILE	"/etc/pam_per_user.map"
#define PAM_PER_USER_DATA_NAME		"pam_per_user_subrequest_handle"
#define PAM_PER_USER_BUFSIZE		1024

#ifndef LOG_AUTHPRIV
# define LOG_AUTHPRIV			LOG_AUTH
#endif


/******************************************************************************
***  prototype for condition-checking functions
******************************************************************************/

/*
** results from condition-checking functions
**
** Note: COND_DEFAULT_MATCH indicates that this condition does match,
**       but only if there are no other, more specific matches
**       later in the map file
*/
enum condition_results
{
	COND_MATCH		=  0,	/* condition matches */
	COND_NO_MATCH		=  1,	/* condition does not match */
	COND_DEFAULT_MATCH	=  2,	/* condition matches by default */
	COND_ERROR		= -1	/* error */
};

/*
** prototype for condition checking function
** arguments:
**	type-dependant key from map file
**	user name
*/
typedef enum condition_results (*cond_func_t)(char *, char *);


/******************************************************************************
***  USER condition function
******************************************************************************/

/*
** condition_user() - check for user name match
*/
static enum condition_results
condition_user(char *mapkey, char *user)
{
#ifdef DEBUG
	printf("==> condition_user(mapkey=\"%s\", user=\"%s\")\n",
	       mapkey, user);
#endif

	if (strcmp(mapkey, "*") == 0)
	{
#ifdef DEBUG
		printf("<== condition_user(): returning COND_DEFAULT_MATCH\n");
#endif
		return COND_DEFAULT_MATCH;
	}

	if (strcmp(mapkey, user) == 0)
	{
#ifdef DEBUG
		printf("<== condition_user(): match\n");
#endif
		return COND_MATCH;
	}

#ifdef DEBUG
	printf("<== condition_user(): no match\n");
#endif
	return COND_NO_MATCH;
}


/******************************************************************************
***  GROUP condition function
******************************************************************************/

/*
** condition_group() - check for group membership of a user
*/
static enum condition_results
condition_group(char *group, char *user)
{
	struct group *grp;
	struct passwd *pwd;
	int i;

#ifdef DEBUG
	printf("==> condition_group(user=\"%s\", group=\"%s\")\n",
	       user, group);
#endif

	/* get group info */
	grp = getgrnam(group);
	if (grp == NULL)
		return COND_ERROR;

	/* loop through group members, see if user is in the list */
	for (i = 0; grp->gr_mem[i] != NULL; i++)
	{
#ifdef DEBUG	
		printf("==> condition_group(): i=%d, member=\"%s\"\n",
		       i, grp->gr_mem[i]);
#endif
 
		if (strcmp(user, grp->gr_mem[i]) == 0)
		{
#ifdef DEBUG
			printf ("<== condition_group(): success\n");
#endif
			return COND_MATCH;
		}
	}

	/* check user's primary gid */
	pwd = getpwnam(user);
	if (pwd == NULL)
		return COND_ERROR;

	/* check whether user's primary gid matches group's gid */
	if (grp->gr_gid == pwd->pw_gid)
		return COND_MATCH;

#ifdef DEBUG
	printf ("<== condition_group(): not found\n");
#endif
	return COND_NO_MATCH;
}


/******************************************************************************
***  list of condition types
******************************************************************************/

/* max length of type name */
#define MAXTYPELEN	32

struct condition_type
{
	char ct_name[MAXTYPELEN];
	cond_func_t ct_func;
};
typedef struct condition_type condition_type_t;

static condition_type_t condition_types[] = {
	{ "GROUP",	condition_group },
	{ "USER",	condition_user },
	{ "",		NULL }
};

/*
** check_condition() - choose a condition function and then call it
*/
static enum condition_results
check_condition(char *cond, char *key, char *user)
{
	int i;

	for (i = 0; condition_types[i].ct_func != NULL; i++)
	{
		if (strcmp(condition_types[i].ct_name, cond) == 0)
			return (*(condition_types[i].ct_func))(key, user);
	}

	/* unknown condition type */
	/* FIXME: syslog a warning message */
	return COND_ERROR;
}


/******************************************************************************
***  map lookup code
******************************************************************************/

/*
** per_user_map_lookup() - look up key in map file
** returns:
**	PAM_SUCCESS		- success
**	PAM_AUTHINFO_UNAVAIL	- error opening map file
**	PAM_AUTH_ERR		- user not found in map file
*/
static int
per_user_map_lookup(char *file, char *user, char *service, size_t servicelen)
{
	FILE *fp;
	char buf[PAM_PER_USER_BUFSIZE];
	char *keyp, *valp, *condp;
	char *cp;

#ifdef DEBUG
	printf("==> per_user_map_lookup(file=\"%s\", user=\"%s\")\n",
	       (file ? file : "[NULL]"), user);
#endif

	/*
	** if this is still an emtpy string at the end, then we know
	** that we haven't found any COND_DEFAULT_MATCH conditions
	*/
	service[0] = '\0';

	fp = fopen(file, "r");
	if (fp == NULL)
	{
		syslog(LOG_AUTHPRIV | LOG_ERR,
		       "pam_per_user: per_user_map_lookup(): "
		       "cannot open mapfile %s: %m",
		       file);
		return PAM_AUTHINFO_UNAVAIL;
	}

	while (fgets(buf, sizeof(buf), fp) != NULL)
	{
		/* strip newline */
		buf[strlen(buf) - 1] = '\0';

#ifdef DEBUG
		printf("    per_user_map_lookup(): buf=\"%s\"\n", buf);
#endif

		/* strip comments */
		if ((cp = strchr(buf, '#')) != NULL
		    && (cp == buf || *(cp - 1) != '\\'))
			*cp = '\0';

		/* strip leading whitespace */
		keyp = buf + strspn(buf, " \t");

		/* skip blank lines */
		if (*keyp == '\0')
			continue;

		/* seperate the key and value */
		valp = NULL;
		if ((cp = strchr(keyp, ':')) != NULL)
		{
			*cp = '\0';
			valp = cp + 1;

			/* strip leading and trailing white space in value */
			valp += strspn(valp, " \t");
			valp[strcspn(valp, " \t")] = '\0';
		}

		/* no value, so skip this entry */
		if (valp == NULL
		    || *valp == '\0')
		{
			/*
			** FIXME: log an "invalid entry on line ###"
			**        warning here
			*/
			continue;
		}

		/* find the key's condition type (if specified) */
		cp = strpbrk(keyp, "=");
		if (cp != NULL)
		{
			*cp++ = '\0';
			condp = keyp;
			keyp = cp;

			/* skip whitespace surrounding "=" character */
			keyp += strspn(keyp, " \t");
			condp[strcspn(condp, " \t")] = '\0';
		}
		else
			/* if none specified, default is "USER" */
			condp = "USER";

		/* strip trailing whitespace in key (if any) */
		if ((cp = strpbrk(keyp, " \t")) != NULL)
			*cp = '\0';

#ifdef DEBUG
		printf("    per_user_map_lookup(): condp=\"%s\", keyp=\"%s\"\n",
		       condp, (keyp ? keyp : "[NULL]"));
#endif

		/*
		** evaluate condition
		*/
		switch (check_condition(condp, keyp, user))
		{
		case COND_MATCH:
			/* condition matches, return specified value */
#ifdef DEBUG
			printf("    per_user_map_lookup(): "
			       "returning valp=\"%s\"\n", valp);
#endif
			strlcpy(service, valp, servicelen);
			fclose(fp);
			return PAM_SUCCESS;

		case COND_DEFAULT_MATCH:
			strlcpy(service, valp, servicelen);
			continue;

		case COND_ERROR:
			/* FIXME: log error */
			continue;

		case COND_NO_MATCH:
			/* no match - continue on to next entry */
			continue;
		}
	}

	fclose(fp);

	/* check whether we encountered a COND_DEFAULT_MATCH earlier */
	if (service[0] != '\0')
	{
#ifdef DEBUG
		printf("<== per_user_map_lookup(): default match\n");
#endif
		return PAM_SUCCESS;
	}

#ifdef DEBUG
	printf("<== per_user_map_lookup(): no match\n");
#endif
	syslog(LOG_AUTHPRIV | LOG_ERR,
	       "pam_per_user: per_user_map_lookup(): "
	       "user %s not found", user);
	return PAM_AUTH_ERR;
}


/******************************************************************************
***  copy PAM items to PAM subrequest handle
******************************************************************************/

static void
copy_pam_item(pam_handle_t *from_pamh, pam_handle_t *to_pamh, int type)
{
	void *item;

#ifdef DEBUG
	printf("=== copy_pam_item(from_pamh=0x%lx, to_pamh=0x%lx, type=%d)\n",
	       from_pamh, to_pamh, type);
#endif

	if (pam_get_item(from_pamh, type, &item) != PAM_SUCCESS)
	{
		syslog(LOG_AUTHPRIV | LOG_ERR,
		       "pam_per_user: copy_pam_item(): pam_get_item(%d) failed",
		       type);
		return;
	}

	if (item == NULL)
		return;

	if (pam_set_item(to_pamh, type, item) != PAM_SUCCESS)
	{
		syslog(LOG_AUTHPRIV | LOG_ERR,
		       "pam_per_user: copy_pam_item(): pam_set_item(%d) failed",
		       type);
		return;
	}
}


static void
copy_pam_items(pam_handle_t *from_pamh, pam_handle_t *to_pamh)
{
	/*
	** Note: We don't copy PAM_SERVICE or PAM_CONV items,
	**       since they should already be set by pam_start()
	*/
	copy_pam_item(from_pamh, to_pamh, PAM_USER);
	copy_pam_item(from_pamh, to_pamh, PAM_TTY);
	copy_pam_item(from_pamh, to_pamh, PAM_RHOST);
	copy_pam_item(from_pamh, to_pamh, PAM_AUTHTOK);
	copy_pam_item(from_pamh, to_pamh, PAM_OLDAUTHTOK);
	copy_pam_item(from_pamh, to_pamh, PAM_RUSER);
	copy_pam_item(from_pamh, to_pamh, PAM_USER_PROMPT);
#ifdef PAM_REPOSITORY
	copy_pam_item(from_pamh, to_pamh, PAM_REPOSITORY);
#endif
#ifdef PAM_FAIL_DELAY
	copy_pam_item(from_pamh, to_pamh, PAM_FAIL_DELAY);
#endif
}


/******************************************************************************
***  allocate and cleanup the PAM subrequest handle
******************************************************************************/

static void
cleanup_subrequest_handle(pam_handle_t *pamh, void *data, int error_status)
{
	pam_end((pam_handle_t *)data, error_status);
}


/*
** create_subrequest_handle() - do map lookup and (if needed) create new
**				subrequest handle
** Note: *sub_pamh will be NULL if there is no prior subrequest
** Return values:
**	PAM_SUCCESS (*sub_pamh set)	- subreq handle successfully created
**	PAM_SUCCESS (*sub_pamh NULL)	- "@SUCCEED" entry
**	PAM_AUTH_ERR			- "@FAIL" entry
**	PAM_IGNORE			- "@IGNORE" entry
**	PAM_AUTHINFO_UNAVAIL		- could not get PAM_CONV item
*/
static int
create_subrequest_handle(pam_handle_t *pamh, char *user, char *mapfile,
			 pam_handle_t **sub_pamh)
{
	char service[PAM_PER_USER_BUFSIZE];
	char *old_service = "";
	struct pam_conv *conv;
	int i;

	syslog(LOG_AUTHPRIV | LOG_INFO,
	       "pam_per_user: create_subrequest_handle(): "
	       "doing map lookup for user \"%s\"",
	       user);

	if (mapfile == NULL)
		mapfile = PAM_PER_USER_DEFAULT_MAPFILE;

	i = per_user_map_lookup(mapfile, user,
				service, sizeof(service));
	if (i != PAM_SUCCESS)
		return i;

#ifdef DEBUG
	printf("    create_subrequest_handle(): service=\"%s\"\n", service);
#endif

	if (strcmp(service, "@IGNORE") == 0)
		return PAM_IGNORE;
	if (strcmp(service, "@FAIL") == 0)
		return PAM_AUTH_ERR;
	if (strcmp(service, "@SUCCEED") == 0)
	{
		/*
		** need to make sure *sub_pamh is NULL, since it might
		** be non-NULL if we switched from a user with a service
		** name to a user with an "@SUCCEED" entry
		*/
		*sub_pamh = NULL;
		return PAM_SUCCESS;
	}

	/*
	** optimization: if the existing subrequest handle is for the
	** same service, then we reuse it instead of creating a new one
	** (note: we'll copy PAM_USER from the parent later on)
	*/
	if (*sub_pamh != NULL)
	{
		if (pam_get_item(*sub_pamh, PAM_SERVICE,
				 (void **)&old_service) != PAM_SUCCESS)
		{
			syslog(LOG_AUTHPRIV | LOG_WARNING,
			       "pam_per_user: create_subrequest_handle(): "
			       "pam_get_item(subreq, PAM_SERVICE) failed");

			/*
			** don't need to do anything explicit here; just
			** let the strcmp() call below fail
			*/
		}

		if (strcmp(old_service, service) == 0)
			return PAM_SUCCESS;
	}

	syslog(LOG_AUTHPRIV | LOG_INFO,
	       "pam_per_user: create_subrequest_handle(): "
	       "creating new subrequest (user=\"%s\", service=\"%s\")",
	       user, service);

	if (pam_get_item(pamh, PAM_CONV, (void **)&conv) != PAM_SUCCESS)
	{
		syslog(LOG_AUTHPRIV | LOG_ERR,
		       "pam_per_user: create_subrequest_handle(): "
		       "pam_get_item(PAM_CONV) failed");
		return PAM_AUTHINFO_UNAVAIL;
	}

	/* create a subrequest */
	if (pam_start(service, NULL, conv, sub_pamh) != PAM_SUCCESS)
	{
		syslog(LOG_AUTHPRIV | LOG_ERR,
		       "pam_per_user: create_subrequest_handle(): "
		       "cannot create subrequest");
		return PAM_AUTH_ERR;
	}

	/* save subrequest handle for later cleanup */
	pam_set_data(pamh, PAM_PER_USER_DATA_NAME, *sub_pamh,
		     cleanup_subrequest_handle);

	return PAM_SUCCESS;
}


static int
get_subrequest_handle(char *mapfile, pam_handle_t *pamh,
		      pam_handle_t **sub_pamh)
{
	char *user, *sub_user = "";
	int retval = PAM_SUCCESS;

#ifdef DEBUG
	printf("==> get_subrequest_handle(mapfile=\"%s\", "
	       "pamh=0x%lx, sub_pamh=0x%lx)\n",
	       (mapfile == NULL ? "[NULL]" : mapfile), pamh, sub_pamh);
#endif

	*sub_pamh = NULL;

	/*
	** get the user name
	*/
	if (pam_get_user(pamh, &user, "login: ") != PAM_SUCCESS)
	{
		syslog(LOG_AUTHPRIV | LOG_ERR,
		       "pam_per_user: get_subrequest_handle(): "
		       "pam_get_user() failed");
		return PAM_USER_UNKNOWN;
	}

#ifdef DEBUG
	printf("    get_subrequest_handle(): user=\"%s\"\n", user);
#endif

	/*
	** check for cached subrequest handle
	*/
	if (pam_get_data(pamh, PAM_PER_USER_DATA_NAME,
			 (const void **)sub_pamh) == PAM_SUCCESS)
	{
		/*
		** if username of subrequest is different from our
		** username, we need to do another map lookup
		*/
		if (pam_get_item(*sub_pamh, PAM_USER,
				 (void **)&sub_user) != PAM_SUCCESS)
		{
			syslog(LOG_AUTHPRIV | LOG_WARNING,
			       "pam_per_user: get_subrequest_handle(): "
			       "pam_get_item(subreq, PAM_USER) failed");

			/*
			** don't need to do anything explicit here; just
			** let the strcmp() call below fail
			*/
		}

		if (strcmp(sub_user, user) != 0)
		{
			retval = create_subrequest_handle(pamh, user,
							  mapfile, sub_pamh);
		}
	}
	else	/* no cached subrequest handle */
	{
		/* note: *sub_pamh is NULL here */
		retval = create_subrequest_handle(pamh, user,
						  mapfile, sub_pamh);
	}

	/*
	** this will be true iff we have a subrequest handle
	** (i.e., it will not be true if there was an error, or if
	** we saw "@FAIL", "@SUCCEED", or "@IGNORE")
	*/
	if (retval == PAM_SUCCESS && *sub_pamh != NULL)
	{
		/*
		** copy PAM items from parent
		**
		** NOTE: We do this every time we're called, since some
		** applications change various items between calls.  The
		** worst example of this is /bin/login, which can change
		** the user name between calls!  This causes a security
		** hole if we don't make sure things are in sync.
		**
		** FIXME: should we copy these items from the
		** subrequest handle back to the parent once we're done?
		*/
		copy_pam_items(pamh, *sub_pamh);
	}

#ifdef DEBUG
	printf("<== get_subrequest_handle(): success\n");
#endif

	return retval;
}


/******************************************************************************
***  public functions
******************************************************************************/

int
pam_sm_authenticate(pam_handle_t *pamh, int flags,
		    int argc, const char *argv[])
{
	pam_handle_t *sub_pamh;
	int i;

#ifdef DEBUG
	printf("==> pam_per_user_authenticate(pamh=0x%lx, flags=%d)\n",
	       pamh, flags);
#endif

	i = get_subrequest_handle((argc == 1 ? (char *)argv[0] : NULL),
				  pamh, &sub_pamh);
#ifdef DEBUG
	printf("    pam_per_user_authenticate(): get_subrequest_handle() returned %d\n", i);
#endif
	if (i != PAM_SUCCESS || sub_pamh == NULL)
		return i;

#ifdef DEBUG
	printf("    pam_per_user_authenticate(): sub_pamh=0x%lx\n", sub_pamh);
#endif

	return pam_authenticate(sub_pamh, flags);
}


int
pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char *argv[])
{
	pam_handle_t *sub_pamh;
	int i;

	i = get_subrequest_handle((argc == 1 ? (char *)argv[0] : NULL),
				  pamh, &sub_pamh);
	if (i != PAM_SUCCESS || sub_pamh == NULL)
		return i;

	return pam_setcred(sub_pamh, flags);
}


int
pam_sm_acct_mgmt(pam_handle_t *pamh, int flags, int argc, const char *argv[])
{
	pam_handle_t *sub_pamh;
	int i;

	i = get_subrequest_handle((argc == 1 ? (char *)argv[0] : NULL),
				  pamh, &sub_pamh);
	if (i != PAM_SUCCESS || sub_pamh == NULL)
		return i;

	return pam_acct_mgmt(sub_pamh, flags);
}


int
pam_sm_open_session(pam_handle_t *pamh, int flags,
		    int argc, const char *argv[])
{
	pam_handle_t *sub_pamh;
	int i;

	i = get_subrequest_handle((argc == 1 ? (char *)argv[0] : NULL),
				  pamh, &sub_pamh);
	if (i != PAM_SUCCESS || sub_pamh == NULL)
		return i;

	return pam_open_session(sub_pamh, flags);
}


int
pam_sm_close_session(pam_handle_t *pamh, int flags,
		     int argc, const char *argv[])
{
	pam_handle_t *sub_pamh;
	int i;

	i = get_subrequest_handle((argc == 1 ? (char *)argv[0] : NULL),
				  pamh, &sub_pamh);
	if (i != PAM_SUCCESS || sub_pamh == NULL)
		return i;

	return pam_close_session(sub_pamh, flags);
}


int
pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, const char *argv[])
{
	pam_handle_t *sub_pamh;
	int i;

	i = get_subrequest_handle((argc == 1 ? (char *)argv[0] : NULL),
				  pamh, &sub_pamh);
	if (i != PAM_SUCCESS || sub_pamh == NULL)
		return i;

	return pam_chauthtok(sub_pamh, flags);
}


